@@ -4,31 +4,32 @@ Adapters for date/time types.
 
 # Copyright (C) 2020 The Psycopg Team
 
-import re
-import sys
-import struct
-from datetime import date, datetime, time, timedelta, timezone
-from typing import Any, Callable, cast, Optional, Tuple, TYPE_CHECKING
+from __future__ import annotations
 
-from .. import postgres
+import re
+import struct
+from typing import TYPE_CHECKING, Any, Callable, cast
+from datetime import date, datetime, time, timedelta, timezone
+
+from .. import _oids
 from ..pq import Format
 from .._tz import get_tzinfo
 from ..abc import AdaptContext, DumperKey
 from ..adapt import Buffer, Dumper, Loader, PyFormat
-from ..errors import InterfaceError, DataError
+from ..errors import DataError, InterfaceError
 from .._struct import pack_int4, pack_int8, unpack_int4, unpack_int8
 
 if TYPE_CHECKING:
-    from ..connection import BaseConnection
+    from .._connection_base import BaseConnection
 
 _struct_timetz = struct.Struct("!qi")  # microseconds, sec tz offset
 _pack_timetz = cast(Callable[[int, int], bytes], _struct_timetz.pack)
-_unpack_timetz = cast(Callable[[bytes], Tuple[int, int]], _struct_timetz.unpack)
+_unpack_timetz = cast(Callable[[Buffer], "tuple[int, int]"], _struct_timetz.unpack)
 
 _struct_interval = struct.Struct("!qii")  # microseconds, days, months
 _pack_interval = cast(Callable[[int, int, int], bytes], _struct_interval.pack)
 _unpack_interval = cast(
-    Callable[[bytes], Tuple[int, int, int]], _struct_interval.unpack
+    Callable[[Buffer], "tuple[int, int, int]"], _struct_interval.unpack
 )
 
 utc = timezone.utc
@@ -39,21 +40,19 @@ _py_date_min_days = date.min.toordinal()
 
 
 class DateDumper(Dumper):
+    oid = _oids.DATE_OID
 
-    oid = postgres.types["date"].oid
-
-    def dump(self, obj: date) -> bytes:
+    def dump(self, obj: date) -> Buffer | None:
         # NOTE: whatever the PostgreSQL DateStyle input format (DMY, MDY, YMD)
         # the YYYY-MM-DD is always understood correctly.
         return str(obj).encode()
 
 
 class DateBinaryDumper(Dumper):
-
     format = Format.BINARY
-    oid = postgres.types["date"].oid
+    oid = _oids.DATE_OID
 
-    def dump(self, obj: date) -> bytes:
+    def dump(self, obj: date) -> Buffer | None:
         days = obj.toordinal() - _pg_date_epoch_days
         return pack_int4(days)
 
@@ -70,15 +69,21 @@ class _BaseTimeDumper(Dumper):
     def upgrade(self, obj: time, format: PyFormat) -> Dumper:
         raise NotImplementedError
 
+    def _get_offset(self, obj: time) -> timedelta:
+        if (offset := obj.utcoffset()) is None:
+            raise DataError(
+                f"cannot calculate the offset of tzinfo '{obj.tzinfo}' without a date"
+            )
+        return offset
+
 
 class _BaseTimeTextDumper(_BaseTimeDumper):
-    def dump(self, obj: time) -> bytes:
+    def dump(self, obj: time) -> Buffer | None:
         return str(obj).encode()
 
 
 class TimeDumper(_BaseTimeTextDumper):
-
-    oid = postgres.types["time"].oid
+    oid = _oids.TIME_OID
 
     def upgrade(self, obj: time, format: PyFormat) -> Dumper:
         if not obj.tzinfo:
@@ -88,16 +93,18 @@ class TimeDumper(_BaseTimeTextDumper):
 
 
 class TimeTzDumper(_BaseTimeTextDumper):
+    oid = _oids.TIMETZ_OID
 
-    oid = postgres.types["timetz"].oid
+    def dump(self, obj: time) -> Buffer | None:
+        self._get_offset(obj)
+        return super().dump(obj)
 
 
 class TimeBinaryDumper(_BaseTimeDumper):
-
     format = Format.BINARY
-    oid = postgres.types["time"].oid
+    oid = _oids.TIME_OID
 
-    def dump(self, obj: time) -> bytes:
+    def dump(self, obj: time) -> Buffer | None:
         us = obj.microsecond + 1_000_000 * (
             obj.second + 60 * (obj.minute + 60 * obj.hour)
         )
@@ -111,16 +118,14 @@ class TimeBinaryDumper(_BaseTimeDumper):
 
 
 class TimeTzBinaryDumper(_BaseTimeDumper):
-
     format = Format.BINARY
-    oid = postgres.types["timetz"].oid
+    oid = _oids.TIMETZ_OID
 
-    def dump(self, obj: time) -> bytes:
+    def dump(self, obj: time) -> Buffer | None:
         us = obj.microsecond + 1_000_000 * (
             obj.second + 60 * (obj.minute + 60 * obj.hour)
         )
-        off = obj.utcoffset()
-        assert off is not None
+        off = self._get_offset(obj)
         return _pack_timetz(us, -int(off.total_seconds()))
 
 
@@ -138,15 +143,14 @@ class _BaseDatetimeDumper(Dumper):
 
 
 class _BaseDatetimeTextDumper(_BaseDatetimeDumper):
-    def dump(self, obj: datetime) -> bytes:
+    def dump(self, obj: datetime) -> Buffer | None:
         # NOTE: whatever the PostgreSQL DateStyle input format (DMY, MDY, YMD)
         # the YYYY-MM-DD is always understood correctly.
         return str(obj).encode()
 
 
 class DatetimeDumper(_BaseDatetimeTextDumper):
-
-    oid = postgres.types["timestamptz"].oid
+    oid = _oids.TIMESTAMPTZ_OID
 
     def upgrade(self, obj: datetime, format: PyFormat) -> Dumper:
         if obj.tzinfo:
@@ -156,16 +160,14 @@ class DatetimeDumper(_BaseDatetimeTextDumper):
 
 
 class DatetimeNoTzDumper(_BaseDatetimeTextDumper):
-
-    oid = postgres.types["timestamp"].oid
+    oid = _oids.TIMESTAMP_OID
 
 
 class DatetimeBinaryDumper(_BaseDatetimeDumper):
-
     format = Format.BINARY
-    oid = postgres.types["timestamptz"].oid
+    oid = _oids.TIMESTAMPTZ_OID
 
-    def dump(self, obj: datetime) -> bytes:
+    def dump(self, obj: datetime) -> Buffer | None:
         delta = obj - _pg_datetimetz_epoch
         micros = delta.microseconds + 1_000_000 * (86_400 * delta.days + delta.seconds)
         return pack_int8(micros)
@@ -178,35 +180,36 @@ class DatetimeBinaryDumper(_BaseDatetimeDumper):
 
 
 class DatetimeNoTzBinaryDumper(_BaseDatetimeDumper):
-
     format = Format.BINARY
-    oid = postgres.types["timestamp"].oid
+    oid = _oids.TIMESTAMP_OID
 
-    def dump(self, obj: datetime) -> bytes:
+    def dump(self, obj: datetime) -> Buffer | None:
         delta = obj - _pg_datetime_epoch
         micros = delta.microseconds + 1_000_000 * (86_400 * delta.days + delta.seconds)
         return pack_int8(micros)
 
 
 class TimedeltaDumper(Dumper):
+    oid = _oids.INTERVAL_OID
 
-    oid = postgres.types["interval"].oid
-
-    def __init__(self, cls: type, context: Optional[AdaptContext] = None):
+    def __init__(self, cls: type, context: AdaptContext | None = None):
         super().__init__(cls, context)
-        if self.connection:
-            if (
-                self.connection.pgconn.parameter_status(b"IntervalStyle")
-                == b"sql_standard"
-            ):
-                setattr(self, "dump", self._dump_sql)
+        if _get_intervalstyle(self.connection) == b"sql_standard":
+            self._dump_method = self._dump_sql
+        else:
+            self._dump_method = self._dump_any
 
-    def dump(self, obj: timedelta) -> bytes:
+    def dump(self, obj: timedelta) -> Buffer | None:
+        return self._dump_method(self, obj)
+
+    @staticmethod
+    def _dump_any(self: TimedeltaDumper, obj: timedelta) -> bytes:
         # The comma is parsed ok by PostgreSQL but it's not documented
         # and it seems brittle to rely on it. CRDB doesn't consume it well.
         return str(obj).encode().replace(b",", b"")
 
-    def _dump_sql(self, obj: timedelta) -> bytes:
+    @staticmethod
+    def _dump_sql(self: TimedeltaDumper, obj: timedelta) -> bytes:
         # sql_standard format needs explicit signs
         # otherwise -1 day 1 sec will mean -1 sec
         return b"%+d day %+d second %+d microsecond" % (
@@ -217,25 +220,22 @@ class TimedeltaDumper(Dumper):
 
 
 class TimedeltaBinaryDumper(Dumper):
-
     format = Format.BINARY
-    oid = postgres.types["interval"].oid
+    oid = _oids.INTERVAL_OID
 
-    def dump(self, obj: timedelta) -> bytes:
+    def dump(self, obj: timedelta) -> Buffer | None:
         micros = 1_000_000 * obj.seconds + obj.microseconds
         return _pack_interval(micros, obj.days, 0)
 
 
 class DateLoader(Loader):
-
     _ORDER_YMD = 0
     _ORDER_DMY = 1
     _ORDER_MDY = 2
 
-    def __init__(self, oid: int, context: Optional[AdaptContext] = None):
+    def __init__(self, oid: int, context: AdaptContext | None = None):
         super().__init__(oid, context)
-        ds = _get_datestyle(self.connection)
-        if ds.startswith(b"I"):  # ISO
+        if (ds := _get_datestyle(self.connection)).startswith(b"I"):  # ISO
             self._order = self._ORDER_YMD
         elif ds.startswith(b"G"):  # German
             self._order = self._ORDER_DMY
@@ -271,7 +271,6 @@ class DateLoader(Loader):
 
 
 class DateBinaryLoader(Loader):
-
     format = Format.BINARY
 
     def load(self, data: Buffer) -> date:
@@ -286,12 +285,10 @@ class DateBinaryLoader(Loader):
 
 
 class TimeLoader(Loader):
-
     _re_format = re.compile(rb"^(\d+):(\d+):(\d+)(?:\.(\d+))?")
 
     def load(self, data: Buffer) -> time:
-        m = self._re_format.match(data)
-        if not m:
+        if not (m := self._re_format.match(data)):
             s = bytes(data).decode("utf8", "replace")
             raise DataError(f"can't parse time {s!r}")
 
@@ -313,7 +310,6 @@ class TimeLoader(Loader):
 
 
 class TimeBinaryLoader(Loader):
-
     format = Format.BINARY
 
     def load(self, data: Buffer) -> time:
@@ -328,9 +324,6 @@ class TimeBinaryLoader(Loader):
 
 
 class TimetzLoader(Loader):
-
-    _py37 = sys.version_info >= (3, 7)
-
     _re_format = re.compile(
         rb"""(?ix)
         ^
@@ -341,8 +334,7 @@ class TimetzLoader(Loader):
     )
 
     def load(self, data: Buffer) -> time:
-        m = self._re_format.match(data)
-        if not m:
+        if not (m := self._re_format.match(data)):
             s = bytes(data).decode("utf8", "replace")
             raise DataError(f"can't parse timetz {s!r}")
 
@@ -360,7 +352,7 @@ class TimetzLoader(Loader):
         off = 60 * 60 * int(oh)
         if om:
             off += 60 * int(om)
-        if os and self._py37:
+        if os:
             off += int(os)
         tz = timezone(timedelta(0, off if sgn == b"+" else -off))
 
@@ -372,7 +364,6 @@ class TimetzLoader(Loader):
 
 
 class TimetzBinaryLoader(Loader):
-
     format = Format.BINARY
 
     def load(self, data: Buffer) -> time:
@@ -383,25 +374,12 @@ class TimetzBinaryLoader(Loader):
         h, m = divmod(val, 60)
 
         try:
-            return time(h, m, s, us, self._tz_from_sec(off))
+            return time(h, m, s, us, timezone(timedelta(seconds=-off)))
         except ValueError:
             raise DataError(f"time not supported by Python: hour={h}") from None
 
-    def _tz_from_sec(self, sec: int) -> timezone:
-        return timezone(timedelta(seconds=-sec))
-
-    def _tz_from_sec_36(self, sec: int) -> timezone:
-        if sec % 60:
-            sec = round(sec / 60.0) * 60
-        return timezone(timedelta(seconds=-sec))
-
-
-if sys.version_info < (3, 7):
-    setattr(TimetzBinaryLoader, "_tz_from_sec", TimetzBinaryLoader._tz_from_sec_36)
-
 
 class TimestampLoader(Loader):
-
     _re_format = re.compile(
         rb"""(?ix)
         ^
@@ -431,11 +409,9 @@ class TimestampLoader(Loader):
     _ORDER_PGDM = 3
     _ORDER_PGMD = 4
 
-    def __init__(self, oid: int, context: Optional[AdaptContext] = None):
+    def __init__(self, oid: int, context: AdaptContext | None = None):
         super().__init__(oid, context)
-
-        ds = _get_datestyle(self.connection)
-        if ds.startswith(b"I"):  # ISO
+        if (ds := _get_datestyle(self.connection)).startswith(b"I"):  # ISO
             self._order = self._ORDER_YMD
         elif ds.startswith(b"G"):  # German
             self._order = self._ORDER_DMY
@@ -448,8 +424,7 @@ class TimestampLoader(Loader):
             raise InterfaceError(f"unexpected DateStyle: {ds.decode('ascii')}")
 
     def load(self, data: Buffer) -> datetime:
-        m = self._re_format.match(data)
-        if not m:
+        if not (m := self._re_format.match(data)):
             raise _get_timestamp_load_error(self.connection, data) from None
 
         if self._order == self._ORDER_YMD:
@@ -487,7 +462,6 @@ class TimestampLoader(Loader):
 
 
 class TimestampBinaryLoader(Loader):
-
     format = Format.BINARY
 
     def load(self, data: Buffer) -> datetime:
@@ -502,7 +476,6 @@ class TimestampBinaryLoader(Loader):
 
 
 class TimestamptzLoader(Loader):
-
     _re_format = re.compile(
         rb"""(?ix)
         ^
@@ -515,17 +488,21 @@ class TimestamptzLoader(Loader):
         """
     )
 
-    def __init__(self, oid: int, context: Optional[AdaptContext] = None):
+    def __init__(self, oid: int, context: AdaptContext | None = None):
         super().__init__(oid, context)
         self._timezone = get_tzinfo(self.connection.pgconn if self.connection else None)
 
-        ds = _get_datestyle(self.connection)
-        if not ds.startswith(b"I"):  # not ISO
-            setattr(self, "load", self._load_notimpl)
+        if _get_datestyle(self.connection).startswith(b"I"):  # ISO
+            self._load_method = self._load_iso
+        else:
+            self._load_method = self._load_notimpl
 
     def load(self, data: Buffer) -> datetime:
-        m = self._re_format.match(data)
-        if not m:
+        return self._load_method(self, data)
+
+    @staticmethod
+    def _load_iso(self: TimestamptzLoader, data: Buffer) -> datetime:
+        if not (m := self._re_format.match(data)):
             raise _get_timestamp_load_error(self.connection, data) from None
 
         ye, mo, da, ho, mi, se, fr, sgn, oh, om, os = m.groups()
@@ -569,7 +546,8 @@ class TimestamptzLoader(Loader):
 
         raise _get_timestamp_load_error(self.connection, data, ex) from None
 
-    def _load_notimpl(self, data: Buffer) -> datetime:
+    @staticmethod
+    def _load_notimpl(self: TimestamptzLoader, data: Buffer) -> datetime:
         s = bytes(data).decode("utf8", "replace")
         ds = _get_datestyle(self.connection).decode("ascii")
         raise NotImplementedError(
@@ -578,10 +556,9 @@ class TimestamptzLoader(Loader):
 
 
 class TimestamptzBinaryLoader(Loader):
-
     format = Format.BINARY
 
-    def __init__(self, oid: int, context: Optional[AdaptContext] = None):
+    def __init__(self, oid: int, context: AdaptContext | None = None):
         super().__init__(oid, context)
         self._timezone = get_tzinfo(self.connection.pgconn if self.connection else None)
 
@@ -596,10 +573,9 @@ class TimestamptzBinaryLoader(Loader):
             # timezone) we can still save the day by shifting the value by the
             # timezone offset and then replacing the timezone.
             if self._timezone:
-                utcoff = self._timezone.utcoffset(
+                if utcoff := self._timezone.utcoffset(
                     datetime.min if micros < 0 else datetime.max
-                )
-                if utcoff:
+                ):
                     usoff = 1_000_000 * int(utcoff.total_seconds())
                     try:
                         ts = _pg_datetime_epoch + timedelta(microseconds=micros + usoff)
@@ -615,7 +591,6 @@ class TimestamptzBinaryLoader(Loader):
 
 
 class IntervalLoader(Loader):
-
     _re_interval = re.compile(
         rb"""
         (?: ([-+]?\d+) \s+ years? \s* )?                # Years
@@ -627,16 +602,19 @@ class IntervalLoader(Loader):
         re.VERBOSE,
     )
 
-    def __init__(self, oid: int, context: Optional[AdaptContext] = None):
+    def __init__(self, oid: int, context: AdaptContext | None = None):
         super().__init__(oid, context)
-        if self.connection:
-            ints = self.connection.pgconn.parameter_status(b"IntervalStyle")
-            if ints != b"postgres":
-                setattr(self, "load", self._load_notimpl)
+        if _get_intervalstyle(self.connection) == b"postgres":
+            self._load_method = self._load_postgres
+        else:
+            self._load_method = self._load_notimpl
 
     def load(self, data: Buffer) -> timedelta:
-        m = self._re_interval.match(data)
-        if not m:
+        return self._load_method(self, data)
+
+    @staticmethod
+    def _load_postgres(self: IntervalLoader, data: Buffer) -> timedelta:
+        if not (m := self._re_interval.match(data)):
             s = bytes(data).decode("utf8", "replace")
             raise DataError(f"can't parse interval {s!r}")
 
@@ -662,20 +640,16 @@ class IntervalLoader(Loader):
             s = bytes(data).decode("utf8", "replace")
             raise DataError(f"can't parse interval {s!r}: {e}") from None
 
-    def _load_notimpl(self, data: Buffer) -> timedelta:
+    @staticmethod
+    def _load_notimpl(self: IntervalLoader, data: Buffer) -> timedelta:
         s = bytes(data).decode("utf8", "replace")
-        ints = (
-            self.connection
-            and self.connection.pgconn.parameter_status(b"IntervalStyle")
-            or b"unknown"
-        ).decode("utf8", "replace")
+        ints = _get_intervalstyle(self.connection).decode("utf8", "replace")
         raise NotImplementedError(
             f"can't parse interval with IntervalStyle {ints}: {s!r}"
         )
 
 
 class IntervalBinaryLoader(Loader):
-
     format = Format.BINARY
 
     def load(self, data: Buffer) -> timedelta:
@@ -693,17 +667,22 @@ class IntervalBinaryLoader(Loader):
             raise DataError(f"can't parse interval: {e}") from None
 
 
-def _get_datestyle(conn: Optional["BaseConnection[Any]"]) -> bytes:
-    if conn:
-        ds = conn.pgconn.parameter_status(b"DateStyle")
-        if ds:
-            return ds
+def _get_datestyle(conn: BaseConnection[Any] | None) -> bytes:
+    if conn and (ds := conn.pgconn.parameter_status(b"DateStyle")):
+        return ds
 
     return b"ISO, DMY"
 
 
+def _get_intervalstyle(conn: BaseConnection[Any] | None) -> bytes:
+    if conn and (ints := conn.pgconn.parameter_status(b"IntervalStyle")):
+        return ints
+
+    return b"unknown"
+
+
 def _get_timestamp_load_error(
-    conn: Optional["BaseConnection[Any]"], data: Buffer, ex: Optional[Exception] = None
+    conn: BaseConnection[Any] | None, data: Buffer, ex: Exception | None = None
 ) -> Exception:
     s = bytes(data).decode("utf8", "replace")
 
@@ -711,8 +690,7 @@ def _get_timestamp_load_error(
         if not s:
             return False
 
-        ds = _get_datestyle(conn)
-        if not ds.startswith(b"P"):  # Postgres
+        if not _get_datestyle(conn).startswith(b"P"):  # Postgres
             return len(s.split()[0]) > 10  # date is first token
         else:
             return len(s.split()[-1]) > 4  # year is last token
